import os
import tempfile
from datetime import datetime

import django
from django.apps import apps
from django.conf import ENVIRONMENT_VARIABLE, settings
from django.core.cache import caches
from django.utils import timezone

from rosetta.conf import settings as rosetta_settings


cache = caches[rosetta_settings.ROSETTA_CACHE_NAME]


def timestamp_with_timezone(dt=None):
    """
    Return a timestamp with a timezone for the configured locale.  If all else
    fails, consider localtime to be UTC.
    """
    dt = dt or datetime.now()
    if timezone is None:
        return dt.strftime("%Y-%m-%d %H:%M%z")
    if not dt.tzinfo:
        tz = timezone.get_current_timezone()
        if not tz:
            tz = timezone.utc
        dt = dt.replace(tzinfo=timezone.get_current_timezone())
    return dt.strftime("%Y-%m-%d %H:%M%z")


def find_pos(lang, project_apps=True, django_apps=False, third_party_apps=False):
    """
    scans a couple possible repositories of gettext catalogs for the given
    language code

    """

    paths = []

    # project/locale
    if settings.SETTINGS_MODULE:
        parts = settings.SETTINGS_MODULE.split(".")
    else:
        # if settings.SETTINGS_MODULE is None, we are probably in "test" mode
        # and override_settings() was used
        # see: https://code.djangoproject.com/ticket/25911
        parts = os.environ.get(ENVIRONMENT_VARIABLE).split(".")
    project = __import__(parts[0], {}, {}, [])
    abs_project_path = os.path.normpath(
        os.path.abspath(os.path.dirname(project.__file__))
    )
    if project_apps:
        if os.path.exists(
            os.path.abspath(os.path.join(os.path.dirname(project.__file__), "locale"))
        ):
            paths.append(
                os.path.abspath(
                    os.path.join(os.path.dirname(project.__file__), "locale")
                )
            )
        if os.path.exists(
            os.path.abspath(
                os.path.join(os.path.dirname(project.__file__), "..", "locale")
            )
        ):
            paths.append(
                os.path.abspath(
                    os.path.join(os.path.dirname(project.__file__), "..", "locale")
                )
            )

    # is OS case sensitive? settings preferred over auto detection
    case_sensitive_file_system = getattr(
        settings, "ROSETTA_CASE_SENSITIVE_FILESYSTEM", None
    )

    # in case of no settings, attempt auto detection
    if case_sensitive_file_system is None:
        case_sensitive_file_system = True
        tmphandle, tmppath = tempfile.mkstemp()
        if os.path.exists(tmppath.upper()):
            # Case insensitive file system.
            case_sensitive_file_system = False

    # django/locale
    if django_apps:
        django_paths = cache.get("rosetta_django_paths")
        if django_paths is None:
            django_paths = []
            for root, dirnames, filename in os.walk(
                os.path.abspath(os.path.dirname(django.__file__))
            ):
                if "locale" in dirnames:
                    django_paths.append(os.path.join(root, "locale"))
                    continue
            cache.set("rosetta_django_paths", django_paths, 60 * 60)
        paths = paths + django_paths
    # settings
    for localepath in settings.LOCALE_PATHS:
        if os.path.isdir(localepath):
            paths.append(localepath)

    # project/app/locale
    for app_ in apps.get_app_configs():
        if (
            rosetta_settings.EXCLUDED_APPLICATIONS
            and app_.name in rosetta_settings.EXCLUDED_APPLICATIONS
        ):
            continue

        app_path = app_.path
        # django apps
        if "contrib" in app_path and "django" in app_path and not django_apps:
            continue

        # third party external
        if not third_party_apps and abs_project_path not in app_path:
            continue

        # local apps
        if not project_apps and abs_project_path in app_path:
            continue

        if os.path.exists(os.path.abspath(os.path.join(app_path, "locale"))):
            paths.append(os.path.abspath(os.path.join(app_path, "locale")))
        if os.path.exists(os.path.abspath(os.path.join(app_path, "..", "locale"))):
            paths.append(os.path.abspath(os.path.join(app_path, "..", "locale")))

    ret = set()
    langs = [lang]
    if "-" in lang:
        _l, _c = map(lambda x: x.lower(), lang.split("-", 1))
        langs += [
            "%s_%s" % (_l, _c),
            "%s_%s" % (_l, _c.upper()),
            "%s_%s" % (_l, _c.capitalize()),
        ]
    elif "_" in lang:
        _l, _c = map(lambda x: x.lower(), lang.split("_", 1))
        langs += [
            "%s-%s" % (_l, _c),
            "%s-%s" % (_l, _c.upper()),
            "%s_%s" % (_l, _c.capitalize()),
        ]

    paths = map(os.path.normpath, paths)
    paths = list(set(paths))
    for path in paths:
        # Exclude paths
        if path not in rosetta_settings.ROSETTA_EXCLUDED_PATHS:
            for lang_ in langs:
                dirname = os.path.join(path, lang_, "LC_MESSAGES")
                for fn in rosetta_settings.POFILENAMES:
                    filename = os.path.join(dirname, fn)
                    abs_path = os.path.abspath(filename)
                    # On case insensitive filesystems (looking at you, MacOS)
                    # compare the lowercase absolute path of the po file
                    # to all lowercased paths already collected.
                    # This is not an issue on sane filesystems
                    if not case_sensitive_file_system:
                        if filename.lower() in [p.lower() for p in ret]:
                            continue
                    if os.path.isfile(abs_path):
                        ret.add(abs_path)
    return list(sorted(ret))


def pagination_range(first, last, current):
    r = []

    r.append(first)
    if first + 1 < last:
        r.append(first + 1)

    if current - 2 > first and current - 2 < last:
        r.append(current - 2)
    if current - 1 > first and current - 1 < last:
        r.append(current - 1)
    if current > first and current < last:
        r.append(current)
    if current + 1 < last and current + 1 > first:
        r.append(current + 1)
    if current + 2 < last and current + 2 > first:
        r.append(current + 2)

    if last - 1 > first:
        r.append(last - 1)
    r.append(last)

    r = list(set(r))
    r.sort()
    prev = 10000
    for e in r[:]:
        if prev + 1 < e:
            try:
                r.insert(r.index(e), "...")
            except ValueError:
                pass
        prev = e
    return r



#!/usr/bin/env python
import os
import sys


if __name__ == "__main__":
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "testproject.settings")

    import django

    django.setup()

    from django.core.management import execute_from_command_line

    execute_from_command_line(sys.argv)



import os
import sys


SITE_ID = 1

PROJECT_PATH = os.path.abspath(os.path.dirname(__file__))

PYTHON_VERSION = "%s.%s" % sys.version_info[:2]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(PROJECT_PATH, "rosetta.db"),
    }
}

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.memcached.PyMemcacheCache",
        "LOCATION": "127.0.0.1:11211",
        "KEY_PREFIX": "ROSETTA_TEST",
    }
}

# CACHES = {'default': {'BACKEND': 'django.core.cache.backends.dummy.DummyCache'}}

TEST_DATABASE_CHARSET = "utf8"
TEST_DATABASE_COLLATION = "utf8_general_ci"

DATABASE_SUPPORTS_TRANSACTIONS = True
SETTINGS_MODULE = "testproject.settings"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.admin",
    # 'django.contrib.admin.apps.SimpleAdminConfig',
    # 'django.contrib.redirects.apps.RedirectsConfig',
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.messages",
    "rosetta",
    "rosetta.tests.test_app.apps.TestAppConfig",
]

LANGUAGE_CODE = "en"

MIDDLEWARE = (
    "django.middleware.common.CommonMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
)

# Note: languages are overridden in the test runner
LANGUAGES = (
    ("en", "English"),
    ("bs-Cyrl-BA", "Bosnian (Cyrillic) (Bosnia and Herzegovina)"),
    ("ja", "日本語"),
    ("xx", "XXXXX"),
    ("fr", "French"),
    ("zh_Hans", "Chinese (Simplified)"),
    ("fr_FR.utf8", "French (France), UTF8"),
)


SILENCED_SYSTEM_CHECKS = ["translation.E002"]


LOCALE_PATHS = [os.path.join(PROJECT_PATH, "locale")]

FIXTURE_DIRS = (os.path.join(PROJECT_PATH, "fixtures"),)
STATIC_URL = "/static/"
ROOT_URLCONF = "testproject.urls"

DEBUG = True

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "debug": False,
            "context_processors": (
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.request",
            ),
        },
    }
]

USE_TZ = True

STATIC_URL = "/static/"

# SESSION_ENGINE = "django.contrib.sessions.backends.signed_cookies"
# ROSETTA_STORAGE_CLASS = 'rosetta.storage.SessionRosettaStorage'

ROSETTA_STORAGE_CLASS = "rosetta.storage.CacheRosettaStorage"
SECRET_KEY = "empty"

ROSETTA_ENABLE_REFLANG = True
ROSETTA_ENABLE_TRANSLATION_SUGGESTIONS = True
ROSETTA_SHOW_AT_ADMIN_PANEL = True
# ROSETTA_SHOW_OCCURRENCES = True

# fake azure key that matches the one in
# fixtures/vcr_cassettes/test_47_azure_ajax_translation.yaml
AZURE_CLIENT_SECRET = None  # "FAKE"

DEEPL_AUTH_KEY = None


# Deepl API language codes are different then those of django, so if this is not set according to your desired languages,
# We use the uppercase version of the first 2 letters of django language code.
# In which case it would work fine for most of the languages,
# But for 'en' if you want "EN-GB" for example, please set it in this dictionary.
# Please check the supported languages list of DeepL API: https://www.deepl.com/docs-api/translating-text/request/
DEEPL_LANGUAGES = None  # ex: {"fr": "FR", "en": "EN-GB", "zh_Hans": "ZH"}



import json
import uuid

import requests

from django.conf import settings


class TranslationException(Exception):
    pass


def translate(text, from_language, to_language):
    AZURE_CLIENT_SECRET = getattr(settings, "AZURE_CLIENT_SECRET", None)
    GOOGLE_APPLICATION_CREDENTIALS_PATH = getattr(
        settings, "GOOGLE_APPLICATION_CREDENTIALS_PATH", None
    )
    GOOGLE_PROJECT_ID = getattr(settings, "GOOGLE_PROJECT_ID", None)
    DEEPL_AUTH_KEY = getattr(settings, "DEEPL_AUTH_KEY", None)

    if DEEPL_AUTH_KEY:
        deepl_language_code = None
        DEEPL_LANGUAGES = getattr(settings, "DEEPL_LANGUAGES", None)
        if type(DEEPL_LANGUAGES) is dict:
            deepl_language_code = DEEPL_LANGUAGES.get(to_language, None)

        if deepl_language_code is None:
            deepl_language_code = to_language[:2].upper()

        return translate_by_deepl(
            text,
            deepl_language_code.upper(),
            DEEPL_AUTH_KEY,
        )

    elif AZURE_CLIENT_SECRET:
        return translate_by_azure(text, from_language, to_language, AZURE_CLIENT_SECRET)
    elif GOOGLE_APPLICATION_CREDENTIALS_PATH and GOOGLE_PROJECT_ID:
        return translate_by_google(
            text,
            from_language,
            to_language,
            GOOGLE_APPLICATION_CREDENTIALS_PATH,
            GOOGLE_PROJECT_ID,
        )
    else:
        raise TranslationException("No translation API service is configured.")


def translate_by_deepl(text, to_language, auth_key):
    if auth_key.lower().endswith(":fx"):
        endpoint = "https://api-free.deepl.com"
    else:
        endpoint = "https://api.deepl.com"

    r = requests.post(
        f"{endpoint}/v2/translate",
        headers={"Authorization": f"DeepL-Auth-Key {auth_key}"},
        data={
            "target_lang": to_language.upper(),
            "text": text,
        },
    )
    return r.json().get("translations")[0].get("text")


def translate_by_azure(text, from_language, to_language, subscription_key):
    """
    This method does the heavy lifting of connecting to the translator API and fetching a response
    :param text: The source text to be translated
    :param from_language: The language of the source text
    :param to_language: The target language to translate the text into
    :param subscription_key: An API key that grants you access to the Azure translation service
    :return: Returns the response from the AZURE service as a python object. For more information about the
    response, please visit
    https://docs.microsoft.com/en-us/azure/cognitive-services/translator/reference/v3-0-translate?tabs=curl
    """

    AZURE_TRANSLATOR_HOST = "https://api.cognitive.microsofttranslator.com"
    AZURE_TRANSLATOR_PATH = "/translate?api-version=3.0"

    headers = {
        "Ocp-Apim-Subscription-Key": subscription_key,
        "Content-type": "application/json",
        "X-ClientTraceId": str(uuid.uuid4()),
    }

    url_parameters = {"from": from_language, "to": to_language}

    request_data = [{"text": text}]

    api_hostname = AZURE_TRANSLATOR_HOST + AZURE_TRANSLATOR_PATH
    r = requests.post(
        api_hostname,
        headers=headers,
        params=url_parameters,
        data=json.dumps(request_data),
    )
    api_response = json.loads(r.text)

    try:
        # result will be a dict if there is an error, e.g.
        # {
        #   "success": false,
        #    "error": "Microsoft Translation API error: Error code 401000,
        #             The request is not authorized because credentials are missing or invalid."
        # }
        if isinstance(api_response, dict):
            api_error = api_response.get("error")
            error_code = api_error.get("code")
            error_message = api_error.get("message")

            raise TranslationException(
                "Microsoft Translation API error: Error code {}, {}".format(
                    error_code, error_message
                )
            )
        else:
            # response body will be of the form:
            # [
            #     {
            #         "translations":[
            #             {"text": "some chinese text that gave a build error on travis ci", "to": "zh-Hans"}
            #         ]
            #     }
            # ]
            # for more information, please visit
            # https://docs.microsoft.com/en-us/azure/cognitive-services/translator/reference/v3-0-translate?tabs=curl

            translations = api_response[0].get("translations")
            translated_text = translations[0].get("text")

            return translated_text
    except requests.exceptions.RequestException as err:
        raise TranslationException(
            "Error connecting to Microsoft Translation Service: {0}".format(err)
        )


def translate_by_google(
    text, input_language, output_language, creadentials_path, project_id
):
    from google.cloud import translate as google_translate

    client = google_translate.TranslationServiceClient.from_service_account_json(
        creadentials_path
    )
    parent = "projects/{}/locations/{}".format(project_id, "global")
    try:
        api_response = client.translate_text(
            request=dict(
                parent=parent,
                contents=[text],
                mime_type="text/plain",
                source_language_code=input_language,
                target_language_code=output_language.split(".", 1)[0],
            )
        )
    except Exception as e:
        raise TranslationException("Google API error: {}".format(e))
    else:
        return str(api_response.translations[0].translated_text)



import hashlib
import os
import os.path
import re
import zipfile
from io import BytesIO
from urllib.parse import urlencode

from polib import pofile

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import user_passes_test
from django.core.paginator import Paginator
from django.http import Http404, HttpResponse, HttpResponseRedirect, JsonResponse
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.encoding import force_bytes
from django.utils.functional import Promise, cached_property
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import never_cache
from django.views.generic import TemplateView, View

from . import get_version as get_rosetta_version
from .access import can_translate, can_translate_language
from .conf import settings as rosetta_settings
from .poutil import find_pos, pagination_range, timestamp_with_timezone
from .signals import entry_changed, post_save
from .storage import get_storage
from .translate_utils import TranslationException, translate


def get_app_name(path):
    return path.split("/locale")[0].split("/")[-1]


class LoginURL(Promise):
    """
    Tests friendly login URL, url is resolved at runtime.
    """

    def __str__(self):
        return rosetta_settings.LOGIN_URL


@method_decorator(never_cache, "dispatch")
@method_decorator(
    user_passes_test(lambda user: can_translate(user), LoginURL()), "dispatch"
)
class RosettaBaseMixin(object):
    """A mixin class for Rosetta's class-based views. It provides:
    * security (see class decorators)
    * a property for the 'po_filter' url argument
    """

    def dispatch(self, *args, **kwargs):
        return super(RosettaBaseMixin, self).dispatch(*args, **kwargs)

    @cached_property
    def po_filter(self):
        """Return the filter applied to all of the .po files under consideration
        to determine which file is currently being translated. Options are:
        'all', 'django', 'third-party', 'project'.

        If the filter isn't in this list, throw a 404.
        """
        po_filter = self.kwargs.get("po_filter")
        if po_filter not in {"all", "django", "third-party", "project"}:
            raise Http404
        return po_filter


class RosettaFileLevelMixin(RosettaBaseMixin):
    """Mixin for dealing with views that work specifically with a single
    .po file. In addition to what the super class brings, it adds the following
    properties:
    * language_id (e.g. 'fr'); derived from url, and validated
    * po_file_path (filesystem path to catalog)
    * po_file (pofile object)
    * po_file_is_writable (bool: do we have filesystem write perms to file)
    """

    def _request_request(self, key, default=None):
        if key in self.request.GET:
            return self.request.GET.get(key)
        elif key in self.request.POST:
            return self.request.POST.get(key)
        return default

    @cached_property
    def language_id(self):
        """Determine/return the language id from the url kwargs, after
        validating that:
        1. the language is in rosetta_settings.ROSETTA_LANGUAGES, and
        2. the current user is permitted to translate that language

        (If either of the above fail, throw a 404.)
        """
        # (Formerly known as "rosetta_i18n_lang_code")
        lang_id = self.kwargs["lang_id"]
        if lang_id not in {lang[0] for lang in rosetta_settings.ROSETTA_LANGUAGES}:
            raise Http404
        if not can_translate_language(self.request.user, lang_id):
            raise Http404
        return lang_id

    @cached_property
    def po_file_path(self):
        """Based on the url kwargs, infer and return the path to the .po file to
        be shown/updated.

        Throw a 404 if a file isn't found.
        """
        # This was formerly referred to as 'rosetta_i18n_fn'
        idx = self.kwargs["idx"]
        idx = int(idx)  # idx matched url re expression; calling int() is safe

        third_party_apps = self.po_filter in ("all", "third-party")
        django_apps = self.po_filter in ("all", "django")
        project_apps = self.po_filter in ("all", "project")

        po_paths = find_pos(
            self.language_id,
            project_apps=project_apps,
            django_apps=django_apps,
            third_party_apps=third_party_apps,
        )
        po_paths.sort(key=get_app_name)

        try:
            path = po_paths[idx]
        except IndexError:
            raise Http404
        return path

    @cached_property
    def po_file(self):
        """Return the parsed .po file that is currently being translated/viewed.

        (Note that this parsing also involves marking up each entry with a hash
        of its contents.)
        """
        if self.po_file_is_writable:
            # If we can write changes to file, then we pull it up fresh with
            # each request.
            # XXX: brittle; what if this path doesn't exist? Isn't a .po file?
            po_file = pofile(
                self.po_file_path, wrapwidth=rosetta_settings.POFILE_WRAP_WIDTH
            )
            for entry in po_file:
                # Entry is an object representing a single entry in the catalog.
                # We iterate through the *entire catalog*, pasting a hashed
                # value of the meat of each entry on its side in an attribute
                # called "md5hash".
                str_to_hash = (
                    str(entry.msgid) + str(entry.msgstr) + str(entry.msgctxt or "")
                ).encode("utf8")
                entry.md5hash = hashlib.md5(str_to_hash).hexdigest()
        else:
            storage = get_storage(self.request)
            po_file = storage.get(self.po_file_cache_key, None)
            if not po_file:
                po_file = pofile(self.po_file_path)
                for entry in po_file:
                    # Entry is an object representing a single entry in the
                    # catalog. We iterate through the entire catalog, pasting
                    # a hashed value of the meat of each entry on its side in
                    # an attribute called "md5hash".
                    str_to_hash = (
                        str(entry.msgid) + str(entry.msgstr) + str(entry.msgctxt or "")
                    ).encode("utf8")
                    entry.md5hash = hashlib.new("md5", str_to_hash).hexdigest()
                storage.set(self.po_file_cache_key, po_file)
        return po_file

    @cached_property
    def po_file_cache_key(self):
        """Return the cache key used to save/access the .po file (when actually
        persisted in cache).
        """
        return "po-file-%s" % self.po_file_path

    @cached_property
    def po_file_is_writable(self):
        """Return True if we're able (in terms of file system permissions) to
        write out changes to the .po file we're translating.
        """
        # (This was formerly called 'rosetta_i18n_write'.)
        return os.access(self.po_file_path, os.W_OK)


class TranslationFileListView(RosettaBaseMixin, TemplateView):
    """Lists the languages, the gettext catalog files that can be translated,
    and their translation progress for a filtered list of apps/projects.
    """

    http_method_names = ["get"]
    template_name = "rosetta/file-list.html"

    def get_context_data(self, **kwargs):
        context = super(TranslationFileListView, self).get_context_data(**kwargs)

        third_party_apps = self.po_filter in ("all", "third-party")
        django_apps = self.po_filter in ("all", "django")
        project_apps = self.po_filter in ("all", "project")

        languages = []
        has_pos = False
        for language in rosetta_settings.ROSETTA_LANGUAGES:
            if not can_translate_language(self.request.user, language[0]):
                continue

            po_paths = find_pos(
                language[0],
                project_apps=project_apps,
                django_apps=django_apps,
                third_party_apps=third_party_apps,
            )
            po_files = [
                (get_app_name(lang), os.path.realpath(lang), pofile(lang))
                for lang in po_paths
            ]
            po_files.sort(key=lambda app: app[0])
            languages.append((language[0], _(language[1]), po_files))
            has_pos = has_pos or bool(po_paths)

        context["version"] = get_rosetta_version()
        context["languages"] = languages
        context["has_pos"] = has_pos
        context["po_filter"] = self.po_filter
        return context


class TranslationFormView(RosettaFileLevelMixin, TemplateView):
    """Show a form with a page's worth of messages to be translated; handle its
    submission by updating cached pofile and, if possible, writing out changes
    to existing .po file.

    Query strings that affect what's shown:
    * msg_filter: filters which messages are displayed. One of 'all', 'fuzzy',
      'translated', and 'untranslated'
    * ref_lang: specifies which language should be shown as the source. Only
      applicable when REF_LANG setting is set to True
    * page: which page (number) should be shown of the paginated results (with
      msg_filter or query applied)
    * query: a search string, where only matches are shown. Fields that are
      searched include: source, translated text, "occurence" file path, or
      context hints.
    """

    # Note: due to the unorthodox nature of the form itself, we're not using
    # Django's generic FormView as our base class.
    http_method_names = ["get", "post"]
    template_name = "rosetta/form.html"

    def fix_nls(self, in_, out_):
        """Fixes submitted translations by filtering carriage returns and pairing
        newlines at the begging and end of the translated string with the original
        """
        if 0 == len(in_) or 0 == len(out_):
            return out_

        if "\r" in out_ and "\r" not in in_:
            out_ = out_.replace("\r", "")

        if "\n" == in_[0] and "\n" != out_[0]:
            out_ = "\n" + out_
        elif "\n" != in_[0] and "\n" == out_[0]:
            out_ = out_.lstrip()
        if 0 == len(out_):
            pass
        elif "\n" == in_[-1] and "\n" != out_[-1]:
            out_ = out_ + "\n"
        elif "\n" != in_[-1] and "\n" == out_[-1]:
            out_ = out_.rstrip()
        return out_

    def post(self, request, *args, **kwargs):
        """The only circumstances when we POST is to submit the main form, both
        updating translations (if any changed) and advancing to the next page of
        messages.

        There is no notion of validation of this content; as implemented, unknown
        fields are ignored and a generic failure message is shown.

        Submitted changes are saved out to the specified .po file on the
        filesystem if that file is writable, otherwise the cached version of the
        file is updated (so it can be downloaded). Then the user is redirected
        to the next page of messages (if there is one; otherwise they're
        redirected back to the current page).
        """
        # The message text inputs are captured as hashes of their initial
        # contents, preceded by "m_". Messages with plurals end with their
        # variation number.
        single_text_input_regex = re.compile(r"^m_([0-9a-f]+)$")
        plural_text_input_regex = re.compile(r"^m_([0-9a-f]+)_([0-9]+)$")
        file_change = False
        for field_name, new_msgstr in request.POST.items():
            md5hash = None

            if plural_text_input_regex.match(field_name):
                md5hash, plural_id = plural_text_input_regex.match(field_name).groups()
                md5hash = str(md5hash)
                # polib parses .po files into unicode strings, but
                # doesn't bother to convert plural indexes to int,
                # so we need unicode here.
                plural_id = str(plural_id)

                # Above no longer true as of Polib 1.0.4
                if plural_id and plural_id.isdigit():
                    plural_id = int(plural_id)

            elif single_text_input_regex.match(field_name):
                md5hash = str(single_text_input_regex.match(field_name).groups()[0])
                plural_id = None

            if md5hash is not None:  # Empty string should be processed!
                entry = self.po_file.find(md5hash, "md5hash")
                # If someone did a makemessage, some entries might
                # have been removed, so we need to check.
                if entry:
                    old_msgstr = entry.msgstr
                    if plural_id is not None:  # 0 is ok!
                        entry.msgstr_plural[plural_id] = self.fix_nls(
                            entry.msgid_plural, new_msgstr
                        )
                    else:
                        entry.msgstr = self.fix_nls(entry.msgid, new_msgstr)

                    is_fuzzy = bool(self.request.POST.get("f_%s" % md5hash, False))
                    old_fuzzy = "fuzzy" in entry.flags

                    if old_fuzzy and not is_fuzzy:
                        entry.flags.remove("fuzzy")
                    elif not old_fuzzy and is_fuzzy:
                        entry.flags.append("fuzzy")

                    file_change = True

                    if old_msgstr != new_msgstr or old_fuzzy != is_fuzzy:
                        entry_changed.send(
                            sender=entry,
                            user=request.user,
                            old_msgstr=old_msgstr,
                            old_fuzzy=old_fuzzy,
                            pofile=self.po_file_path,
                            language_code=self.language_id,
                        )
                else:
                    messages.error(
                        self.request,
                        _(
                            "Some items in your last translation block couldn't "
                            "be saved: this usually happens when the catalog file "
                            "changes on disk after you last loaded it."
                        ),
                    )

        if file_change and self.po_file_is_writable:
            try:
                self.po_file.metadata["Last-Translator"] = "{} {} <{}>".format(
                    getattr(self.request.user, "first_name", "Anonymous"),
                    getattr(self.request.user, "last_name", "User"),
                    getattr(self.request.user, "email", "anonymous@user.tld"),
                )
                self.po_file.metadata["X-Translated-Using"] = "django-rosetta %s" % (
                    get_rosetta_version()
                )
                self.po_file.metadata["PO-Revision-Date"] = timestamp_with_timezone()
            except UnicodeDecodeError:
                pass

            try:
                self.po_file.save()
                po_filepath, ext = os.path.splitext(self.po_file_path)

                if rosetta_settings.AUTO_COMPILE:
                    self.po_file.save_as_mofile(po_filepath + ".mo")

                post_save.send(
                    sender=None, language_code=self.language_id, request=self.request
                )
                # Try auto-reloading via the WSGI daemon mode reload mechanism
                should_try_wsgi_reload = (
                    rosetta_settings.WSGI_AUTO_RELOAD
                    and "mod_wsgi.process_group" in self.request.environ
                    and self.request.environ.get("mod_wsgi.process_group", None)
                    and "SCRIPT_FILENAME" in self.request.environ
                    and int(self.request.environ.get("mod_wsgi.script_reloading", 0))
                )
                if should_try_wsgi_reload:
                    try:
                        os.utime(self.request.environ.get("SCRIPT_FILENAME"), None)
                    except OSError:
                        pass
                # Try auto-reloading via uwsgi daemon reload mechanism
                if rosetta_settings.UWSGI_AUTO_RELOAD:
                    try:
                        import uwsgi

                        uwsgi.reload()  # pretty easy right?
                    except Exception:
                        pass  # we may not be running under uwsgi :P
                # XXX: It would be nice to add a success message here!
            except Exception as e:
                messages.error(self.request, e)

        if file_change and not self.po_file_is_writable:
            storage = get_storage(self.request)
            storage.set(self.po_file_cache_key, self.po_file)

        # Reconstitute url to redirect to. Start with determining whether the
        # page number can be incremented.
        paginator = Paginator(self.get_entries(), rosetta_settings.MESSAGES_PER_PAGE)
        try:
            page = int(self._request_request("page", 1))
        except ValueError:
            page = 1  # fall back to page 1
        else:
            if not (0 < page <= paginator.num_pages):
                page = 1
        if page < paginator.num_pages:
            page += 1
        query_string_args = {
            "msg_filter": self.msg_filter,
            "query": self.query,
            "ref_lang": self.ref_lang,
            "page": page,
        }
        # Winnow down the query string args to non-blank ones
        query_string_args = {k: v for k, v in query_string_args.items() if v}
        return HttpResponseRedirect(
            "{url}?{qs}".format(
                url=reverse("rosetta-form", kwargs=self.kwargs),
                qs=urlencode_safe(query_string_args),
            )
        )

    def get_context_data(self, **kwargs):
        context = super(TranslationFormView, self).get_context_data(**kwargs)
        entries = self.get_entries()
        paginator = Paginator(entries, rosetta_settings.MESSAGES_PER_PAGE)

        # Handle REF_LANG setting; mark up our entries with the reg lang's
        # corresponding translations
        LANGUAGES = list(rosetta_settings.ROSETTA_LANGUAGES)
        if rosetta_settings.ENABLE_REFLANG:
            if self.ref_lang_po_file:
                for o in paginator.object_list:
                    ref_entry = self.ref_lang_po_file.find(o.msgid)
                    if ref_entry and ref_entry.msgstr:
                        o.ref_txt = ref_entry.msgstr
                    else:
                        o.ref_txt = o.msgid
            else:
                for o in paginator.object_list:
                    o.ref_txt = o.msgid
            # XXX: having "MSGID" at the end of the dropdown is really odd, no?
            # Why not instead do this?
            # LANGUAGES = [('', '----')] + list(settings.LANGUAGES)
            LANGUAGES.append(("msgid", "MSGID"))

        # Determine page number & how pagination links should be displayed
        try:
            page = int(self._request_request("page", 1))
        except ValueError:
            page = 1  # fall back to page 1
        else:
            if not (0 < page <= paginator.num_pages):
                page = 1
        needs_pagination = paginator.num_pages > 1
        if needs_pagination:
            if paginator.num_pages >= 10:
                page_range = pagination_range(1, paginator.num_pages, page)
            else:
                page_range = range(1, 1 + paginator.num_pages)

        rosetta_messages = paginator.page(page).object_list

        # Handle MAIN_LANGUAGE setting, if applicable; mark up each entry
        # in the pagination window with the "main language"'s string.
        main_language_id = rosetta_settings.MAIN_LANGUAGE
        main_language = None
        if main_language_id and main_language_id != self.language_id:
            # Translate from id to language name
            for language in rosetta_settings.ROSETTA_LANGUAGES:
                if language[0] == main_language_id:
                    main_language = _(language[1])
                    break
        if main_language:
            main_lang_po_path = self.po_file_path.replace(
                "/%s/" % self.language_id, "/%s/" % main_language_id
            )
            # XXX: brittle; what if this path doesn't exist? Isn't a .po file?
            main_lang_po = pofile(main_lang_po_path)

            for message in rosetta_messages:
                message.main_lang = main_lang_po.find(message.msgid).msgstr

        # Collect some constants for the template
        rosetta_i18n_lang_name = str(
            dict(rosetta_settings.ROSETTA_LANGUAGES).get(self.language_id)
        )
        # "bidi" as in "bi-directional"
        rosetta_i18n_lang_bidi = (
            self.language_id.split("-")[0] in settings.LANGUAGES_BIDI
        )
        query_string_args = {}
        if self.msg_filter:
            query_string_args["msg_filter"] = self.msg_filter
        if self.query:
            query_string_args["query"] = self.query
        if self.ref_lang:
            query_string_args["ref_lang"] = self.ref_lang
        # Base for pagination links; the page num itself is added in template
        pagination_query_string_base = urlencode_safe(query_string_args)
        # Base for msg filter links; it doesn't make sense to persist page
        # numbers in these links. We just pass in ref_lang, if it's set.
        filter_query_string_base = urlencode_safe(
            {k: v for k, v in query_string_args.items() if k == "ref_lang"}
        )

        context.update(
            {
                "version": get_rosetta_version(),
                "LANGUAGES": LANGUAGES,
                "rosetta_settings": rosetta_settings,
                "rosetta_i18n_lang_name": rosetta_i18n_lang_name,
                "rosetta_i18n_lang_code": self.language_id,
                "rosetta_i18n_lang_code_normalized": self.language_id.replace("_", "-"),
                "rosetta_i18n_lang_bidi": rosetta_i18n_lang_bidi,
                "rosetta_i18n_filter": self.msg_filter,
                "rosetta_i18n_write": self.po_file_is_writable,
                "rosetta_messages": rosetta_messages,
                "page_range": needs_pagination and page_range,
                "needs_pagination": needs_pagination,
                "main_language": main_language,
                "rosetta_i18n_app": get_app_name(self.po_file_path),
                "page": page,
                "query": self.query,
                "pagination_query_string_base": pagination_query_string_base,
                "filter_query_string_base": filter_query_string_base,
                "paginator": paginator,
                "rosetta_i18n_pofile": self.po_file,
                "ref_lang": self.ref_lang,
            }
        )

        return context

    @cached_property
    def ref_lang(self):
        """Return the language id for the "reference language" (the language to
        be translated *from*, if not English).

        Throw a 404 if it's not in rosetta_settings.ROSETTA_LANGUAGES.
        """
        ref_lang = self._request_request("ref_lang", "msgid")
        if ref_lang != "msgid":
            allowed_languages = {lang[0] for lang in rosetta_settings.ROSETTA_LANGUAGES}
            if ref_lang not in allowed_languages:
                raise Http404
        return ref_lang

    @cached_property
    def ref_lang_po_file(self):
        """Return a parsed .po file object for the "reference language", if one
        exists, otherwise None.
        """
        ref_pofile = None
        if rosetta_settings.ENABLE_REFLANG and self.ref_lang != "msgid":
            replacement = "{separator}locale{separator}{ref_lang}".format(
                separator=os.sep, ref_lang=self.ref_lang
            )
            pattern = r"\{separator}locale\{separator}[a-z]{{2}}".format(
                separator=os.sep
            )
            ref_fn = re.sub(pattern, replacement, self.po_file_path)
            try:
                ref_pofile = pofile(ref_fn)
            except IOError:
                # there's a syntax error in the PO file and polib can't
                # open it. Let's just do nothing and thus display msgids.
                # XXX: :-/
                pass
        return ref_pofile

    @cached_property
    def msg_filter(self):
        """Validate/return msg_filter from request (e.g. 'fuzzy', 'untranslated'),
        or a default.

        If a query is also specified in the request, then return None.
        """
        if self.query:
            msg_filter = None
        else:
            msg_filter = self._request_request("msg_filter", "all")
            available_msg_filters = {"untranslated", "translated", "fuzzy", "all"}
            if msg_filter not in available_msg_filters:
                msg_filter = "all"
        return msg_filter

    @cached_property
    def query(self):
        """Strip and return the query (for searching the catalog) from the
        request, or None.
        """
        return self._request_request("query", "").strip() or None

    def get_entries(self):
        """Return a list of the entries (messages) that would be part of the
        current "view"; that is, all of the ones from this .po file matching the
        current query or msg_filter.
        """
        if self.query:
            # Scenario #1: terms matching a search query
            rx = re.compile(re.escape(self.query), re.IGNORECASE)

            def concat_entry(e):
                return (
                    str(e.msgstr)
                    + str(e.msgid)
                    + str(e.msgctxt)
                    + str(e.comment)
                    + "".join([o[0] for o in e.occurrences])
                    + str(e.msgid_plural)
                    + "".join(e.msgstr_plural.values())
                )

            entries = [
                e_
                for e_ in self.po_file
                if not e_.obsolete and rx.search(concat_entry(e_))
            ]
        else:
            # Scenario #2: filtered list of messages
            if self.msg_filter == "untranslated":
                entries = self.po_file.untranslated_entries()
            elif self.msg_filter == "translated":
                entries = self.po_file.translated_entries()
            elif self.msg_filter == "fuzzy":
                entries = [e_ for e_ in self.po_file.fuzzy_entries() if not e_.obsolete]
            else:
                # ("all")
                entries = [e_ for e_ in self.po_file if not e_.obsolete]
        return entries


class TranslationFileDownload(RosettaFileLevelMixin, View):
    """Download a zip file for a specific catalog including both the raw (.po)
    and compiled (.mo) files, either as they exist on disk, or, if what's on
    disk is unwritable (permissions-wise), return what's in the cache.
    """

    http_method_names = ["get"]

    def get(self, request, *args, **kwargs):
        try:
            if len(self.po_file_path.split("/")) >= 5:
                offered_fn = "_".join(self.po_file_path.split("/")[-5:])
            else:
                offered_fn = self.po_file_path.split("/")[-1]
            po_fn = str(self.po_file_path.split("/")[-1])
            mo_fn = str(po_fn.replace(".po", ".mo"))  # not so smart, huh
            zipdata = BytesIO()
            with zipfile.ZipFile(zipdata, mode="w") as zipf:
                zipf.writestr(po_fn, str(self.po_file).encode("utf8"))
                zipf.writestr(mo_fn, self.po_file.to_binary())
            zipdata.seek(0)

            response = HttpResponse(zipdata.read())
            filename = "filename=%s.%s.zip" % (offered_fn, self.language_id)
            response["Content-Disposition"] = "attachment; %s" % filename
            response["Content-Type"] = "application/x-zip"
            return response
        except Exception:
            # XXX: should add a message!
            return HttpResponseRedirect(
                reverse("rosetta-file-list", kwargs={"po_filter": "project"})
            )


@user_passes_test(lambda user: can_translate(user), LoginURL())
def translate_text(request):

    language_from = request.GET.get("from", None)
    language_to = request.GET.get("to", None)
    text = request.GET.get("text", None)

    if language_from == language_to:
        data = {"success": True, "translation": text}
    else:
        try:
            translated_text = translate(text, language_from, language_to)

            data = {"success": True, "translation": translated_text}
        except TranslationException as e:
            data = {"success": False, "error": str(e)}

    return JsonResponse(data)


def urlencode_safe(query):
    return urlencode({k: force_bytes(v) for k, v in query.items()})



import re

from django import template
from django.utils.html import escape
from django.utils.safestring import mark_safe

from rosetta.access import can_translate


register = template.Library()
rx = re.compile(r"(%(\([^\s\)]*\))?[sd]|\{[\w\d_]+?\})")

can_translate = register.filter(can_translate)


def format_message(message):
    return mark_safe(
        rx.sub("<code>\\1</code>", escape(message).replace(r"\n", "<br />\n"))
    )


format_message = register.filter(format_message)


def lines_count(message):
    return 1 + sum([len(line) / 50 for line in message.split("\n")])


lines_count = register.filter(lines_count)


def mult(a, b):
    return int(a) * int(b)


mult = register.filter(mult)


def minus(a, b):
    try:
        return int(a) - int(b)
    except Exception:
        return 0


minus = register.filter(minus)


def gt(a, b):
    try:
        return int(a) > int(b)
    except Exception:
        return False


gt = register.filter(gt)


def do_incr(parser, token):
    args = token.split_contents()
    if len(args) < 2:
        raise SyntaxError("'incr' tag requires at least one argument")
    name = args[1]
    if not hasattr(parser, "_namedIncrNodes"):
        parser._namedIncrNodes = {}
    if name not in parser._namedIncrNodes:
        parser._namedIncrNodes[name] = IncrNode(0)
    return parser._namedIncrNodes[name]


do_incr = register.tag("increment", do_incr)


class IncrNode(template.Node):
    def __init__(self, init_val=0):
        self.val = init_val

    def render(self, context):
        self.val += 1
        return str(self.val)


def is_fuzzy(message):
    return message and hasattr(message, "flags") and "fuzzy" in message.flags


is_fuzzy = register.filter(is_fuzzy)



import hashlib
import importlib
import time

from django.conf import settings
from django.core.cache import caches
from django.core.exceptions import ImproperlyConfigured

from .conf import settings as rosetta_settings


cache = caches[rosetta_settings.ROSETTA_CACHE_NAME]


class BaseRosettaStorage(object):
    def __init__(self, request):
        self.request = request

    def get(self, key, default=None):
        raise NotImplementedError

    def set(self, key, val):
        raise NotImplementedError

    def has(self, key):
        raise NotImplementedError

    def delete(self, key):
        raise NotImplementedError


class DummyRosettaStorage(BaseRosettaStorage):
    def get(self, key, default=None):
        return default

    def set(self, key, val):
        pass

    def has(self, key):
        return False

    def delete(self, key):
        pass


class SessionRosettaStorage(BaseRosettaStorage):
    def __init__(self, request):
        super(SessionRosettaStorage, self).__init__(request)

        if (
            "signed_cookies" in settings.SESSION_ENGINE
            and "pickle" not in settings.SESSION_SERIALIZER.lower()
        ):
            raise ImproperlyConfigured(
                "Sorry, but django-rosetta doesn't support the `signed_cookies` SESSION_ENGINE, because rosetta specific session files cannot be serialized."
            )

    def get(self, key, default=None):
        if key in self.request.session:
            return self.request.session[key]
        return default

    def set(self, key, val):
        self.request.session[key] = val

    def has(self, key):
        return key in self.request.session

    def delete(self, key):
        del self.request.session[key]


class CacheRosettaStorage(BaseRosettaStorage):
    # unlike the session storage backend, cache is shared among all users
    # so we need to per-user key prefix, which we store in the session
    def __init__(self, request):
        super(CacheRosettaStorage, self).__init__(request)

        if "rosetta_cache_storage_key_prefix" in self.request.session:
            self._key_prefix = self.request.session["rosetta_cache_storage_key_prefix"]
        else:
            self._key_prefix = hashlib.new(
                "sha1", str(time.time()).encode("utf8")
            ).hexdigest()
            self.request.session["rosetta_cache_storage_key_prefix"] = self._key_prefix

        if self.request.session["rosetta_cache_storage_key_prefix"] != self._key_prefix:
            raise ImproperlyConfigured(
                "You can't use the CacheRosettaStorage because your Django Session storage doesn't seem to be working. The CacheRosettaStorage relies on the Django Session storage to avoid conflicts."
            )

        # Make sure we're not using DummyCache
        if (
            "dummycache"
            in settings.CACHES[rosetta_settings.ROSETTA_CACHE_NAME]["BACKEND"].lower()
        ):
            raise ImproperlyConfigured(
                "You can't use the CacheRosettaStorage if your cache isn't correctly set up (you are using the DummyCache cache backend)."
            )

        # Make sure the cache actually works
        try:
            self.set("rosetta_cache_test", "rosetta")
            if not self.get("rosetta_cache_test") == "rosetta":
                raise ImproperlyConfigured(
                    "You can't use the CacheRosettaStorage if your cache isn't correctly set up, please double check your Django DATABASES setting and that the cache server is responding."
                )
        finally:
            self.delete("rosetta_cache_test")

    def get(self, key, default=None):
        # print ('get', self._key_prefix + key)
        return cache.get(self._key_prefix + key, default)

    def set(self, key, val):
        # print ('set', self._key_prefix + key)
        cache.set(self._key_prefix + key, val, 86400)

    def has(self, key):
        # print ('has', self._key_prefix + key)
        return (self._key_prefix + key) in cache

    def delete(self, key):
        # print ('del', self._key_prefix + key)
        cache.delete(self._key_prefix + key)


def get_storage(request):
    from rosetta.conf import settings

    storage_module, storage_class = settings.STORAGE_CLASS.rsplit(".", 1)
    storage_module = importlib.import_module(storage_module)
    return getattr(storage_module, storage_class)(request)



"""
For a description of each rosetta setting see: docs/settings.rst.
"""

from django.conf import settings as dj_settings
from django.core.signals import setting_changed


__all__ = ["settings"]


class RosettaSettings(object):
    """
    Class that holds rosetta settings.

    The settings object is an instance of this class and is reloaded when
    the ``setting_changed`` signal is dispatched.
    """

    SETTINGS = {
        "ROSETTA_MESSAGES_PER_PAGE": ("MESSAGES_PER_PAGE", 10),
        "ROSETTA_ENABLE_TRANSLATION_SUGGESTIONS": (
            "ENABLE_TRANSLATION_SUGGESTIONS",
            False,
        ),
        "YANDEX_TRANSLATE_KEY": ("YANDEX_TRANSLATE_KEY", None),
        "AZURE_CLIENT_SECRET": ("AZURE_CLIENT_SECRET", None),
        "GOOGLE_APPLICATION_CREDENTIALS_PATH": (
            "GOOGLE_APPLICATION_CREDENTIALS_PATH",
            None,
        ),
        "GOOGLE_PROJECT_ID": ("GOOGLE_PROJECT_ID", None),
        "DEEPL_AUTH_KEY": ("DEEPL_AUTH_KEY", None),
        "ROSETTA_MAIN_LANGUAGE": ("MAIN_LANGUAGE", None),
        "ROSETTA_MESSAGES_SOURCE_LANGUAGE_CODE": (
            "MESSAGES_SOURCE_LANGUAGE_CODE",
            "en",
        ),
        "ROSETTA_MESSAGES_SOURCE_LANGUAGE_NAME": (
            "MESSAGES_SOURCE_LANGUAGE_NAME",
            "English",
        ),
        "ROSETTA_ACCESS_CONTROL_FUNCTION": ("ACCESS_CONTROL_FUNCTION", None),
        "ROSETTA_WSGI_AUTO_RELOAD": ("WSGI_AUTO_RELOAD", False),
        "ROSETTA_UWSGI_AUTO_RELOAD": ("UWSGI_AUTO_RELOAD", False),
        "ROSETTA_EXCLUDED_APPLICATIONS": ("EXCLUDED_APPLICATIONS", ()),
        "ROSETTA_POFILE_WRAP_WIDTH": ("POFILE_WRAP_WIDTH", 78),
        "ROSETTA_STORAGE_CLASS": (
            "STORAGE_CLASS",
            "rosetta.storage.CacheRosettaStorage",
        ),
        "ROSETTA_ENABLE_REFLANG": ("ENABLE_REFLANG", False),
        "ROSETTA_POFILENAMES": ("POFILENAMES", ("django.po", "djangojs.po")),
        "ROSETTA_CACHE_NAME": (
            "ROSETTA_CACHE_NAME",
            "rosetta" if "rosetta" in dj_settings.CACHES else "default",
        ),
        "ROSETTA_REQUIRES_AUTH": ("ROSETTA_REQUIRES_AUTH", True),
        "ROSETTA_EXCLUDED_PATHS": ("ROSETTA_EXCLUDED_PATHS", ()),
        "ROSETTA_LANGUAGE_GROUPS": ("ROSETTA_LANGUAGE_GROUPS", False),
        "ROSETTA_AUTO_COMPILE": ("AUTO_COMPILE", True),
        "ROSETTA_SHOW_AT_ADMIN_PANEL": ("SHOW_AT_ADMIN_PANEL", False),
        "ROSETTA_LOGIN_URL": ("LOGIN_URL", dj_settings.LOGIN_URL),
        "ROSETTA_LANGUAGES": ("ROSETTA_LANGUAGES", dj_settings.LANGUAGES),
        "ROSETTA_SHOW_OCCURRENCES": ("SHOW_OCCURRENCES", True),
        # Deepl API language codes are different then those of django, so if this is not set according to your desired languages,
        # We use the first 2 letters of django language code.
        # In which case it would work fine for most of the languages,
        # But for 'en' if you want "EN-GB" for example, please set it in this dictionary.
        # you can find the supported languages list of DeepL API here: https://www.deepl.com/docs-api/translating-text/request/
        # ex: DEEPL_LANGUAGES = {"fr": "FR", "en": "EN-GB", "zh_Hans": "ZH"}
        "DEEPL_LANGUAGES": ("DEEPL_LANGUAGES", {}),
    }

    def __init__(self):
        # make sure we don't assign self._settings directly here, to avoid
        # recursion in __setattr__, we delegate to the parent instead
        super(RosettaSettings, self).__setattr__("_settings", {})
        self.load()

    def load(self):
        for user_setting, (rosetta_setting, default) in self.SETTINGS.items():
            self._settings[rosetta_setting] = getattr(
                dj_settings, user_setting, default
            )

    def reload(self):
        self.__init__()

    def __getattr__(self, attr):
        if attr not in self._settings:
            raise AttributeError(
                "'RosettaSettings' object has not attribute '%s'" % attr
            )
        return self._settings[attr]

    def __setattr__(self, attr, value):
        if attr not in self._settings:
            raise AttributeError(
                "'RosettaSettings' object has not attribute '%s'" % attr
            )
        self._settings[attr] = value


# This is our global settings object
settings = RosettaSettings()


# Signal handler to reload settings when needed
def reload_settings(*args, **kwargs):
    val = kwargs.get("setting")
    if val in settings.SETTINGS:
        settings.reload()


# Connect the setting_changed signal to our handler
setting_changed.connect(reload_settings)



from django.contrib import admin
from django.contrib.staticfiles.urls import staticfiles_urlpatterns
from django.urls import include, re_path


admin.autodiscover()

urlpatterns = [
    re_path(r"^admin/", admin.site.urls),
    re_path(r"^rosetta/", include("rosetta.urls")),
]

urlpatterns += staticfiles_urlpatterns()



from django.urls import re_path, reverse_lazy
from django.views.generic.base import RedirectView

from . import views


urlpatterns = [
    re_path(
        r"^$",
        RedirectView.as_view(
            url=reverse_lazy("rosetta-file-list", kwargs={"po_filter": "project"}),
            permanent=False,
        ),
        name="rosetta-old-home-redirect",
    ),
    re_path(
        r"^files/$",
        RedirectView.as_view(
            url=reverse_lazy("rosetta-file-list", kwargs={"po_filter": "project"}),
            permanent=False,
        ),
        name="rosetta-file-list-redirect",
    ),
    re_path(
        r"^files/(?P<po_filter>[\w-]+)/$",
        views.TranslationFileListView.as_view(),
        name="rosetta-file-list",
    ),
    re_path(
        r"^files/(?P<po_filter>[\w-]+)/(?P<lang_id>[\w\-_\.@]+)/(?P<idx>\d+)/$",
        views.TranslationFormView.as_view(),
        name="rosetta-form",
    ),
    re_path(
        r"^files/(?P<po_filter>[\w-]+)/(?P<lang_id>[\w\-_\.@]+)/(?P<idx>\d+)/download/$",
        views.TranslationFileDownload.as_view(),
        name="rosetta-download-file",
    ),
    re_path(r"^translate/$", views.translate_text, name="rosetta.translate_text"),
]



from django import dispatch


# providing_args=["user", "old_msgstr", "old_fuzzy", "pofile", "language_code"]
entry_changed = dispatch.Signal()

# providing_args=["language_code", "request"]
post_save = dispatch.Signal()



import filecmp
import hashlib
import os
import shutil
from unittest import mock
from urllib.parse import urlencode

import vcr

from django import VERSION
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.dispatch import receiver
from django.http import Http404
from django.test import RequestFactory, TestCase, override_settings
from django.test.client import Client
from django.urls import resolve, reverse
from django.utils.encoding import force_bytes

from rosetta import views
from rosetta.poutil import find_pos
from rosetta.signals import entry_changed, post_save
from rosetta.storage import get_storage


class RosettaTestCase(TestCase):
    def __init__(self, *args, **kwargs):
        super(RosettaTestCase, self).__init__(*args, **kwargs)
        self.curdir = os.path.dirname(__file__)
        self.dest_file = os.path.normpath(
            os.path.join(self.curdir, "../locale/xx/LC_MESSAGES/django.po")
        )

    def setUp(self):
        from django.contrib.auth.models import User

        user = User.objects.create_superuser(
            "test_admin", "test@test.com", "test_password"
        )
        user2 = User.objects.create_superuser(
            "test_admin2", "test@test2.com", "test_password"
        )
        user3 = User.objects.create_superuser(
            "test_admin3", "test@test2.com", "test_password"
        )

        user3.is_staff = False
        user3.save()

        self.user = user
        self.client2 = Client()

        self.client.login(username=user.username, password="test_password")
        self.client2.login(username=user2.username, password="test_password")

        shutil.copy(self.dest_file, self.dest_file + ".orig")

    def tearDown(self):
        shutil.move(self.dest_file + ".orig", self.dest_file)

    def copy_po_file_from_template(self, template_path):
        """Utility method to handle swapping a template po file in place for
        testing.
        """
        src_path = os.path.normpath(os.path.join(self.curdir, template_path))
        shutil.copy(src_path, self.dest_file)

    @property
    def xx_form_url(self):
        kwargs = {"po_filter": "third-party", "lang_id": "xx", "idx": 0}
        return reverse("rosetta-form", kwargs=kwargs)

    @property
    def all_file_list_url(self):
        return reverse("rosetta-file-list", kwargs={"po_filter": "all"})

    @property
    def project_file_list_url(self):
        return reverse("rosetta-file-list", kwargs={"po_filter": "project"})

    @property
    def third_party_file_list_url(self):
        return reverse("rosetta-file-list", kwargs={"po_filter": "third-party"})

    def test_1_ListLoading(self):
        r = self.client.get(self.third_party_file_list_url)
        self.assertTrue(
            os.path.normpath("rosetta/locale/xx/LC_MESSAGES/django.po")
            in r.content.decode()
        )

    @override_settings(ROSETTA_LANGUAGES=(("xx", "dummy language"),))
    def test_2_PickFile(self):
        r = self.client.get(self.xx_form_url)
        self.assertTrue("dummy language" in r.content.decode())

    def test_3_DownloadZIP(self):
        kwargs = {"po_filter": "third-party", "lang_id": "xx", "idx": 0}
        url = reverse("rosetta-download-file", kwargs=kwargs)
        r = self.client.get(url)
        self.assertTrue("application/x-zip" in r["content-type"])

    @override_settings(ROSETTA_LANGUAGES=(("xx", "dummy language"),))
    def test_4_DoChanges(self):
        self.copy_po_file_from_template("./django.po.template")
        untranslated_url = self.xx_form_url + "?msg_filter=untranslated"
        translated_url = self.xx_form_url + "?msg_filter=translated"

        # Load the template file
        r = self.client.get(untranslated_url)

        # make sure both strings are untranslated
        self.assertTrue("dummy language" in r.content.decode())
        self.assertTrue("String 1" in r.content.decode())
        self.assertTrue("String 2" in r.content.decode())
        self.assertTrue("m_e48f149a8b2e8baa81b816c0edf93890" in r.content.decode())

        # post a translation
        data = {"m_e48f149a8b2e8baa81b816c0edf93890": "Hello, world"}
        r = self.client.post(untranslated_url, data, follow=True)

        # reload all untranslated strings
        r = self.client.get(untranslated_url)

        # the translated string no longer is up for translation
        self.assertTrue("String 1" in r.content.decode())
        self.assertTrue("String 2" not in r.content.decode())

        # display only translated strings
        r = self.client.get(translated_url)

        # The translation was persisted
        self.assertTrue("String 1" not in r.content.decode())
        self.assertTrue("String 2" in r.content.decode())
        self.assertTrue("Hello, world" in r.content.decode())

    @override_settings(ROSETTA_LANGUAGES=(("xx", "dummy language"),))
    def test_5_TestIssue67(self):
        # issue 67: http://code.google.com/p/django-rosetta/issues/detail?id=67
        self.copy_po_file_from_template("./django.po.issue67.template")

        # Make sure the plurals string is valid
        with open(self.dest_file, "r") as f_:
            content = f_.read()
        self.assertTrue("Hello, world" not in content)
        self.assertTrue("|| n%100>=20) ? 1 : 2)" in content)
        del content

        r = self.client.get(self.xx_form_url + "?msg_filter=untranslated")

        # make sure all strings are untranslated
        self.assertTrue("dummy language" in r.content.decode())
        self.assertTrue("String 1" in r.content.decode())
        self.assertTrue("String 2" in r.content.decode())
        self.assertTrue("m_e48f149a8b2e8baa81b816c0edf93890" in r.content.decode())

        # post a translation
        data = {"m_e48f149a8b2e8baa81b816c0edf93890": "Hello, world"}
        self.client.post(self.xx_form_url + "?msg_filter=untranslated", data)

        # Make sure the plurals string is still valid
        with open(self.dest_file, "r") as f_:
            content = f_.read()
        self.assertTrue("Hello, world" in str(content))
        self.assertTrue("|| n%100>=20) ? 1 : 2)" in str(content))
        self.assertTrue("or n%100>=20) ? 1 : 2)" not in str(content))
        del content

    @override_settings(ROSETTA_LANGUAGES=(("xx", "dummy language"),))
    def test_6_ExcludedApps(self):
        with self.settings(ROSETTA_EXCLUDED_APPLICATIONS=("rosetta",)):
            r = self.client.get(self.third_party_file_list_url)
            self.assertNotContains(r, "rosetta/locale/xx/LC_MESSAGES/django.po")

        with self.settings(ROSETTA_EXCLUDED_APPLICATIONS=()):
            r = self.client.get(self.third_party_file_list_url)
            self.assertContains(r, "rosetta/locale/xx/LC_MESSAGES/django.po")

    def test_7_selfInApplist(self):
        r = self.client.get(self.third_party_file_list_url)
        self.assertContains(r, "rosetta/locale/xx/LC_MESSAGES/django.po")

        r = self.client.get(self.project_file_list_url)
        self.assertNotContains(r, "rosetta/locale/xx/LC_MESSAGES/django.po")

    @override_settings(ROSETTA_LANGUAGES=(("xx", "dummy language"),))
    def test_8_hideObsoletes(self):
        r = self.client.get(self.xx_form_url)

        # not in listing
        for p in range(1, 5):
            r = self.client.get(self.xx_form_url + "?page=%d" % p)
            self.assertTrue("dummy language" in r.content.decode())
            self.assertTrue("Les deux" not in r.content.decode())

        r = self.client.get(self.xx_form_url + "?query=Les%20Deux")
        self.assertContains(r, "dummy language")
        self.assertNotContains(r, "Les deux")

    def test_9_concurrency(self):
        self.copy_po_file_from_template("./django.po.template")
        translated_url = self.xx_form_url + "?msg_filter=translated"
        untranslated_url = self.xx_form_url + "?msg_filter=untranslated"

        # Load the template file
        r = self.client.get(untranslated_url)
        r2 = self.client2.get(untranslated_url)

        self.assertContains(r, "String 1")
        self.assertContains(r2, "String 1")
        self.assertContains(r, "m_08e4e11e2243d764fc45a5a4fba5d0f2")

        data = {"m_08e4e11e2243d764fc45a5a4fba5d0f2": "Hello, world"}
        r = self.client.post(untranslated_url, data, follow=True)

        # Client 2 reloads, forces a reload of the catalog; untranslated
        # string1 is now translated
        r2 = self.client2.get(untranslated_url, follow=True)
        self.assertNotContains(r, "String 1")
        self.assertContains(r, "String 2")
        self.assertNotContains(r2, "String 1")
        self.assertContains(r2, "String 2")

        r = self.client.get(untranslated_url)
        r2 = self.client2.get(untranslated_url)

        self.assertContains(r2, "String 2")
        self.assertContains(r2, "m_e48f149a8b2e8baa81b816c0edf93890")
        self.assertContains(r, "String 2")
        self.assertContains(r, "m_e48f149a8b2e8baa81b816c0edf93890")

        # client 2 posts!
        data = {"m_e48f149a8b2e8baa81b816c0edf93890": "Hello, world, from client two!"}
        r2 = self.client2.post(untranslated_url, data, follow=True)

        self.assertNotContains(r2, "save-conflict")

        # uh-oh here comes client 1
        data = {"m_e48f149a8b2e8baa81b816c0edf93890": "Hello, world, from client one!"}
        r = self.client.post(untranslated_url, data, follow=True)
        # An error message is displayed
        self.assertContains(r, "save-conflict")

        # client 2 won
        with open(self.dest_file, "r") as po_file:
            pofile_content = po_file.read()
        self.assertTrue("Hello, world, from client two!" in pofile_content)

        # Both clients show all strings, error messages are gone
        r = self.client.get(translated_url)
        self.assertNotContains(r, "save-conflict")
        r2 = self.client2.get(translated_url)
        self.assertNotContains(r2, "save-conflict")
        r = self.client.get(self.xx_form_url)
        self.assertNotContains(r, "save-conflict")
        r2 = self.client2.get(self.xx_form_url)
        self.assertNotContains(r2, "save-conflict")

        # Both have client's two version
        self.assertContains(r, "Hello, world, from client two!")
        self.assertContains(r2, "Hello, world, from client two!")

    def test_10_issue_79_num_entries(self):
        self.copy_po_file_from_template("./django.po.issue79.template")
        r = self.client.get(self.third_party_file_list_url)
        self.assertContains(r, '<td class="ch-messages r">1</td>')
        self.assertContains(r, '<td class="ch-progress r">0%</td>')
        self.assertContains(r, '<td class="ch-obsolete r">1</td>')

    def test_11_issue_80_tab_indexes(self):
        r = self.client.get(self.xx_form_url)
        self.assertTrue('tabindex="3"' in r.content.decode())

    def test_12_issue_82_staff_user(self):
        self.client3 = Client()
        self.client3.login(username="test_admin3", password="test_password")

        # When auth is required, we get an empty response (and a redirect) with
        # this user.
        with self.settings(ROSETTA_REQUIRES_AUTH=True):
            r = self.client3.get(self.xx_form_url)
            self.assertFalse(r.content.decode())
            self.assertEqual(r.status_code, 302)

        # When it's not required, we sail through.
        with self.settings(ROSETTA_REQUIRES_AUTH=False):
            r = self.client3.get(self.xx_form_url)
            self.assertTrue(r.content.decode())
            self.assertEqual(r.status_code, 200)

    @override_settings(ROSETTA_LANGUAGES=(("fr", "French"), ("xx", "Dummy Language")))
    def test_13_catalog_filters(self):
        r = self.client.get(self.third_party_file_list_url)
        self.assertTrue(
            os.path.normpath("rosetta/locale/xx/LC_MESSAGES/django.po")
            in r.content.decode()
        )

        url = reverse("rosetta-file-list", kwargs={"po_filter": "django"})
        r = self.client.get(url)
        self.assertTrue(
            os.path.normpath("rosetta/locale/xx/LC_MESSAGES/django.po")
            not in r.content.decode()
        )

        r = self.client.get(self.all_file_list_url)
        self.assertTrue(
            os.path.normpath("rosetta/locale/xx/LC_MESSAGES/django.po")
            in r.content.decode()
        )

        r = self.client.get(self.project_file_list_url)
        self.assertTrue(
            os.path.normpath("rosetta/locale/xx/LC_MESSAGES/django.po")
            not in r.content.decode()
        )

    def test_14_issue_99_context_and_comments(self):
        r = self.client.get(self.xx_form_url)
        self.assertTrue("This is a text of the base template" in r.content.decode())
        self.assertTrue("Context hint" in r.content.decode())

    def test_15_issue_87_entry_changed_signal(self):
        self.copy_po_file_from_template("./django.po.template")
        r = self.client.get(self.xx_form_url)

        @receiver(entry_changed)
        def test_receiver(sender, **kwargs):
            self.test_old_msgstr = kwargs.get("old_msgstr")
            self.test_new_msgstr = sender.msgstr
            self.test_msg_id = sender.msgid

        self.assertTrue("m_e48f149a8b2e8baa81b816c0edf93890" in r.content.decode())

        # post a translation
        data = {"m_e48f149a8b2e8baa81b816c0edf93890": "Hello, world"}
        self.client.post(self.xx_form_url, data)

        self.assertTrue(self.test_old_msgstr == "")
        self.assertTrue(self.test_new_msgstr == "Hello, world")
        self.assertTrue(self.test_msg_id == "String 2")

        del (self.test_old_msgstr, self.test_new_msgstr, self.test_msg_id)

    def test_16_issue_101_post_save_signal(self):
        self.copy_po_file_from_template("./django.po.template")
        r = self.client.get(self.xx_form_url)

        @receiver(post_save)
        def test_receiver(sender, **kwargs):
            self.test_sig_lang = kwargs.get("language_code")

        self.assertTrue("m_e48f149a8b2e8baa81b816c0edf93890" in r.content.decode())

        # post a translation
        data = {"m_e48f149a8b2e8baa81b816c0edf93890": "Hello, world"}
        self.client.post(self.xx_form_url, data)

        self.assertTrue(self.test_sig_lang == "xx")
        del self.test_sig_lang

    def test_17_issue_103_post_save_signal_has_request(self):
        self.copy_po_file_from_template("./django.po.template")
        r = self.client.get(self.xx_form_url)

        @receiver(post_save)
        def test_receiver(sender, **kwargs):
            self.test_16_has_request = "request" in kwargs

        self.assertTrue("m_e48f149a8b2e8baa81b816c0edf93890" in r.content.decode())

        # post a translation
        data = {"m_e48f149a8b2e8baa81b816c0edf93890": "Hello, world"}
        r = self.client.post(self.xx_form_url, data)

        self.assertTrue(self.test_16_has_request)
        del self.test_16_has_request

    def test_18_Test_Issue_gh24(self):
        self.copy_po_file_from_template("./django.po.issue24gh.template")
        r = self.client.get(self.xx_form_url)

        self.assertTrue("m_bb9d8fe6159187b9ea494c1b313d23d4" in r.content.decode())

        # Post a translation, it should have properly wrapped lines
        data = {
            "m_bb9d8fe6159187b9ea494c1b313d23d4": (
                "Lorem ipsum dolor sit amet, consectetuer adipiscing elit. Aenean "
                "commodo ligula eget dolor. Aenean massa. Cum sociis natoque "
                "penatibus et magnis dis parturient montes, nascetur ridiculus "
                "mus. Donec quam felis, ultricies nec, pellentesque eu, pretium "
                "quis, sem. Nulla consequat massa quis enim. Donec pede justo, "
                "fringilla vel, aliquet nec, vulputate eget, arcu. In enim justo, "
                "rhoncus ut, imperdiet a, venenatis vitae, justo. Nullam dictum "
                "felis eu pede mollis pretium."
            )
        }
        r = self.client.post(self.xx_form_url, data)
        with open(self.dest_file, "r") as po_file:
            pofile_content = po_file.read()

        self.assertTrue('"pede mollis pretium."' in pofile_content)

        # Again, with unwrapped lines
        self.copy_po_file_from_template("./django.po.issue24gh.template")
        with self.settings(ROSETTA_POFILE_WRAP_WIDTH=0):
            r = self.client.get(self.xx_form_url)
            self.assertTrue("m_bb9d8fe6159187b9ea494c1b313d23d4" in r.content.decode())
            r = self.client.post(self.xx_form_url, data)
            with open(self.dest_file, "r") as po_file:
                pofile_content = po_file.read()
            self.assertTrue('felis eu pede mollis pretium."' in pofile_content)

    def test_19_Test_Issue_gh34(self):
        self.copy_po_file_from_template("./django.po.issue34gh.template")
        r = self.client.get(self.xx_form_url)
        self.assertTrue("m_ff7060c1a9aae9c42af4d54ac8551f67_1" in r.content.decode())
        self.assertTrue("m_ff7060c1a9aae9c42af4d54ac8551f67_0" in r.content.decode())
        self.assertTrue("m_09f7e02f1290be211da707a266f153b3" in r.content.decode())

        # post a translation, it should have properly wrapped lines
        data = {
            "m_ff7060c1a9aae9c42af4d54ac8551f67_0": "Foo %s",
            "m_ff7060c1a9aae9c42af4d54ac8551f67_1": "Bar %s",
            "m_09f7e02f1290be211da707a266f153b3": "Salut",
        }
        r = self.client.post(self.xx_form_url, data)
        with open(self.dest_file, "r") as po_file:
            pofile_content = po_file.read()
        self.assertTrue('msgstr "Salut\\n"' in pofile_content)
        self.assertTrue('msgstr[0] ""\n"\\n"\n"Foo %s\\n"' in pofile_content)
        self.assertTrue('msgstr[1] ""\n"\\n"\n"Bar %s\\n"' in pofile_content)

    @override_settings(
        SESSION_ENGINE="django.contrib.sessions.backends.signed_cookies",
        ROSETTA_STORAGE_CLASS="rosetta.storage.CacheRosettaStorage",
    )
    def test_20_Test_Issue_gh38(self):
        # (Have to log in again, since our session engine changed)
        self.client.login(username="test_admin", password="test_password")
        self.assertTrue(
            "django.contrib.sessions.middleware.SessionMiddleware"
            in settings.MIDDLEWARE
        )

        # Only one backend to test: cache backend
        self.copy_po_file_from_template("./django.po.issue38gh.template")

        r = self.client.get(self.xx_form_url)
        self.assertFalse(len(str(self.client.cookies.get("sessionid"))) > 4096)
        self.assertTrue("m_9efd113f7919952523f06e0d88da9c54" in r.content.decode())

        data = {"m_9efd113f7919952523f06e0d88da9c54": "Testing cookie length"}
        r = self.client.post(self.xx_form_url, data)
        with open(self.dest_file, "r") as po_file:
            pofile_content = po_file.read()
        self.assertTrue("Testing cookie length" in pofile_content)

        r = self.client.get(self.xx_form_url + "?filter=translated")
        self.assertTrue("Testing cookie length" in r.content.decode())
        self.assertTrue("m_9f6c442c6d579707440ba9dada0fb373" in r.content.decode())

    @override_settings(ROSETTA_STORAGE_CLASS="rosetta.storage.CacheRosettaStorage")
    def test_21_concurrency_of_cache_backend(self):
        self.copy_po_file_from_template("./django.po.issue38gh.template")

        # Force caching into play by making .po file read-only
        os.chmod(self.dest_file, 292)  # 0444

        self.client.get(self.xx_form_url)
        self.client2.get(self.xx_form_url)
        self.assertNotEqual(
            self.client.session.get("rosetta_cache_storage_key_prefix"),
            self.client2.session.get("rosetta_cache_storage_key_prefix"),
        )

        # Clean up (restore perms)
        os.chmod(self.dest_file, 420)  # 0644

    def test_22_Test_Issue_gh39(self):
        self.copy_po_file_from_template("./django.po.issue39gh.template")

        r = self.client.get(self.xx_form_url)
        # We have distinct hashes, even though the msgid and msgstr are identical
        self.assertTrue("m_4765f7de94996d3de5975fa797c3451f" in r.content.decode())
        self.assertTrue("m_08e4e11e2243d764fc45a5a4fba5d0f2" in r.content.decode())

    @override_settings(ROSETTA_LANGUAGES=(("xx", "dummy language"),))
    def test_23_save_header_data(self):
        from django.contrib.auth.models import User

        self.copy_po_file_from_template("./django.po.template")

        unicode_user = User.objects.create_user(
            "test_unicode", "save_header_data@test.com", "test_unicode"
        )
        unicode_user.first_name = "aéaéaé aàaàaàa"
        unicode_user.last_name = "aâââ üüüü"
        unicode_user.is_superuser, unicode_user.is_staff = True, True
        unicode_user.save()

        self.client.login(username="test_unicode", password="test_unicode")

        # Load the template file
        r = self.client.get(self.xx_form_url + "?filter=untranslated")

        # make sure both strings are untranslated
        self.assertTrue("dummy language" in r.content.decode())
        self.assertTrue("String 1" in r.content.decode())
        self.assertTrue("String 2" in r.content.decode())
        self.assertTrue("m_e48f149a8b2e8baa81b816c0edf93890" in r.content.decode())

        # post a translation
        data = {"m_e48f149a8b2e8baa81b816c0edf93890": "Hello, world"}
        r = self.client.post(self.xx_form_url + "?filter=untranslated", data)
        # read the result
        with open(self.dest_file, "r") as f_:
            content = f_.read()

        # make sure unicode data was properly converted to ascii
        self.assertTrue("Hello, world" in content)
        self.assertTrue("save_header_data@test.com" in content)
        self.assertTrue("aéaéaé aàaàaàa aâââ üüüü" in content)

    def test_24_percent_translation(self):
        self.copy_po_file_from_template("./django.po.template")

        # Load the template file
        r = self.client.get(self.xx_form_url)

        self.assertTrue("Progress: 0%" in r.content.decode())
        data = {"m_e48f149a8b2e8baa81b816c0edf93890": "Hello, world"}
        r = self.client.post(self.xx_form_url, data, follow=True)
        self.assertTrue("Progress: 25%" in r.content.decode())

    def test_25_replace_access_control(self):
        # Test default access control allows access
        response = self.client.get(self.project_file_list_url)
        self.assertEqual(200, response.status_code)

        # Now replace access control with a function reference,
        # and check we get redirected
        with self.settings(ROSETTA_ACCESS_CONTROL_FUNCTION="rosetta.tests.no_access"):
            response = self.client.get(self.project_file_list_url)
            self.assertEqual(302, response.status_code)

        # Now replace access control with a function itself,
        # and check we get redirected
        with self.settings(ROSETTA_ACCESS_CONTROL_FUNCTION=lambda user: False):
            response = self.client.get(self.project_file_list_url)
            self.assertEqual(302, response.status_code)

    def test_26_urlconf_accept_dots_and_underscores(self):
        resolver_match = resolve("/rosetta/files/all/fr_FR.utf8/0/")
        self.assertEqual(resolver_match.url_name, "rosetta-form")
        self.assertEqual(resolver_match.kwargs["lang_id"], "fr_FR.utf8")

    def test_27_extended_urlconf_language_code_loads_file(self):
        url = reverse(
            "rosetta-form",
            kwargs={"po_filter": "all", "lang_id": "fr_FR.utf8", "idx": 0},
        )
        r = self.client.get(url)
        self.assertTrue("French (France), UTF8" in r.content.decode())
        self.assertTrue("m_03a603523bd75b00414a413657acdeb2" in r.content.decode())

    def test_28_issue_gh87(self):
        """Make sure that rosetta_i18n_catalog_filter is passed into the context."""
        r = self.client.get(self.third_party_file_list_url)
        self.assertContains(
            r, '<li class="active"><a href="/rosetta/files/third-party/">'
        )

    @override_settings(
        SESSION_ENGINE="django.contrib.sessions.backends.signed_cookies",
        ROSETTA_STORAGE_CLASS="rosetta.storage.SessionRosettaStorage",
    )
    def test_29_unsupported_p3_django_16_storage(self):
        if VERSION[0:2] < (2, 0):
            self.assertTrue(
                "django.contrib.sessions.middleware.SessionMiddleware"
                in settings.MIDDLEWARE
            )

            # Force caching to be used by making the pofile read-only
            os.chmod(self.dest_file, 292)  # 0444

            # (Have to log in again, since our session engine changed)
            self.client.login(username="test_admin", password="test_password")

            with self.assertRaises(ImproperlyConfigured):
                self.client.get(self.xx_form_url)

            # Cleanup
            os.chmod(self.dest_file, 420)  # 0644

    @override_settings(
        ROSETTA_POFILENAMES=("pr44.po",), ROSETTA_LANGUAGES=(("xx", "dummy language"),)
    )
    def test_30_pofile_names(self):
        os.unlink(self.dest_file)
        destfile = os.path.normpath(
            os.path.join(self.curdir, "../locale/xx/LC_MESSAGES/pr44.po")
        )
        shutil.copy(
            os.path.normpath(os.path.join(self.curdir, "./pr44.po.template")), destfile
        )

        r = self.client.get(self.third_party_file_list_url)
        self.assertTrue("xx/LC_MESSAGES/pr44.po" in r.content.decode())

        r = self.client.get(self.xx_form_url)
        self.assertTrue("dummy language" in r.content.decode())

        # (Clean up)
        os.unlink(destfile)

    def test_31_pr_102__exclude_paths(self):
        r = self.client.get(self.third_party_file_list_url)
        self.assertContains(r, "rosetta/locale/xx/LC_MESSAGES/django.po")
        exclude_path = os.path.normpath(os.path.join(self.curdir, "../locale"))
        with self.settings(ROSETTA_EXCLUDED_PATHS=exclude_path):
            r = self.client.get(self.third_party_file_list_url)
            self.assertNotContains(r, "rosetta/locale/xx/LC_MESSAGES/django.po")

    def test_32_pr_103__language_groups(self):
        from django.contrib.auth.models import Group, User

        # Default behavior: non-admins need to be in a translators group; they
        # see all catalogs
        translators = Group.objects.create(name="translators")
        translators_xx = Group.objects.create(name="translators-xx")

        user4 = User.objects.create_user(
            "test_admin4", "test@test3.com", "test_password"
        )
        user4.groups.add(translators)
        user4.is_superuser = False
        user4.is_staff = True
        user4.save()

        with self.settings(ROSETTA_LANGUAGE_GROUPS=False):
            self.client.login(username="test_admin4", password="test_password")
            r = self.client.get(self.third_party_file_list_url)
            self.assertContains(r, "rosetta/locale/xx/LC_MESSAGES/django.po")

        with self.settings(ROSETTA_LANGUAGE_GROUPS=True):
            r = self.client.get(self.third_party_file_list_url)
            self.assertNotContains(r, "rosetta/locale/xx/LC_MESSAGES/django.po")
            # Now add them to the custom group
            user4.groups.add(translators_xx)
            r = self.client.get(self.third_party_file_list_url)
            self.assertContains(r, "rosetta/locale/xx/LC_MESSAGES/django.po")

    @override_settings(
        ROSETTA_ENABLE_REFLANG=True, ROSETTA_LANGUAGES=(("xx", "dummy language"),)
    )
    def test_33_reflang(self):
        self.copy_po_file_from_template("./django.po.issue60.template")
        r = self.client.get(self.xx_form_url)

        # Verify that there's an option to select a reflang
        self.assertTrue(
            '<option value="?ref_lang=xx">dummy language</option>' in r.content.decode()
        )

        r = self.client.get(self.xx_form_url + "?ref_lang=xx")
        # The translated string in the test PO file ends up in the "Reference" column
        self.assertTrue(
            '<span class="message">translated-string1</span>' in r.content.decode()
        )

    def test_show_occurrences(self):
        r = self.client.get(self.xx_form_url)
        # Verify that occurrences in view
        self.assertTrue('<td class="location">' in r.content.decode())
        with self.settings(ROSETTA_SHOW_OCCURRENCES=False):
            r = self.client.get(self.xx_form_url)
            # Verify that occurrences not in view
            self.assertFalse('<td class="location">' in r.content.decode())

    def test_34_issue_113_app_configs(self):
        r = self.client.get(self.all_file_list_url)
        self.assertTrue('rosetta/files/all/xx/1/">Test_App' in r.content.decode())

    @override_settings(ROSETTA_STORAGE_CLASS="rosetta.storage.CacheRosettaStorage")
    def test_35_issue_135_display_exception_messages(self):
        # Note: the old version of this test looked for a 'Permission denied'
        # message reflected in the response. That behavior has now changed so
        # that changes that can't be persisted through the filesystem .po file
        # are saved to the cached version of the .po file.
        self.copy_po_file_from_template("./django.po.template")

        r = self.client.get(self.xx_form_url + "?msg_filter=untranslated")
        self.assertContains(r, "m_e48f149a8b2e8baa81b816c0edf93890")

        # make the pofile read-only
        os.chmod(self.dest_file, 292)  # 0444

        # post a translation
        data = {"m_e48f149a8b2e8baa81b816c0edf93890": "Hello, world"}
        self.client.post(self.xx_form_url, data, follow=True)

        # Confirm that the filesystem file hasn't changed
        tmpl_path = os.path.normpath(os.path.join(self.curdir, "django.po.template"))
        self.assertTrue(filecmp.cmp(tmpl_path, self.dest_file))

        # Confirm that the cached version has been updated
        cache_key = "po-file-%s" % self.dest_file
        request = RequestFactory().get(self.xx_form_url)
        request.user = self.user
        request.session = self.client.session
        storage = get_storage(request)

        po_file = storage.get(cache_key)
        entry = po_file.find("String 2")
        self.assertEqual(entry.msgstr, "Hello, world")

        # cleanup
        os.chmod(self.dest_file, 420)  # 0644

    def test_36_issue_142_complex_locales(self):
        r = self.client.get(self.all_file_list_url)
        self.assertContains(r, "locale/bs-Cyrl-BA/LC_MESSAGES/django.po")

    @override_settings(ROSETTA_LANGUAGES=(("yy-Anot", "Yet Another dummy language"),))
    def test_37_issue_133_complex_locales(self):
        r = self.client.get(self.all_file_list_url)
        self.assertContains(r, "locale/yy-Anot/LC_MESSAGES/django.po")

    def test_38_issue_161_more_weird_locales(self):
        r = self.client.get(self.all_file_list_url)
        self.assertTrue(r, "locale/zh_Hans/LC_MESSAGES/django.po")

    def test_39_invalid_get_page(self):
        url = self.xx_form_url + "?filter=untranslated"

        r = self.client.get(url)  # Page not specified
        self.assertEqual(r.context["page"], 1)

        r = self.client.get(url + "&page=")  # No number given
        self.assertEqual(r.context["page"], 1)

        r = self.client.get(url + "&page=9999")  # Too-high number given
        self.assertEqual(r.context["page"], 1)

        r = self.client.get(url + "&page=x")  # Non-number given
        self.assertEqual(r.context["page"], 1)

    def test_40_issue_155_auto_compile(self):
        def file_hash(file_string):
            with open(file_string, encoding="latin-1") as file:
                file_content = file.read().encode("utf-8")
            return hashlib.md5(file_content).hexdigest()

        def message_hashes():
            r = self.client.get(self.xx_form_url)
            return {m.msgid: "m_" + m.md5hash for m in r.context["rosetta_messages"]}

        po_file = self.dest_file
        mo_file = self.dest_file[:-3] + ".mo"

        # MO file will be compiled by default.
        # Get PO and MO files into an initial reference state (MO will be
        # created or updated)
        msg_hashes = message_hashes()
        data = {msg_hashes["String 1"]: "Translation 1"}
        self.client.post(self.xx_form_url, data)
        po_file_hash_before, mo_file_hash_before = (
            file_hash(po_file),
            file_hash(mo_file),
        )

        # Make a change to the translations
        msg_hashes = message_hashes()
        data = {msg_hashes["String 1"]: "Translation 2"}
        self.client.post(self.xx_form_url, data)

        # Get the new hashes of the PO and MO file contents
        po_file_hash_after, mo_file_hash_after = file_hash(po_file), file_hash(mo_file)

        # Both the PO and MO should have changed
        self.assertNotEqual(po_file_hash_before, po_file_hash_after)
        self.assertNotEqual(mo_file_hash_before, mo_file_hash_after)

        # Disable auto-compilation of the MO when the PO is saved
        with self.settings(ROSETTA_AUTO_COMPILE=False):
            # Make a change to the translations
            po_file_hash_before, mo_file_hash_before = (
                po_file_hash_after,
                mo_file_hash_after,
            )
            msg_hashes = message_hashes()
            data = {msg_hashes["String 1"]: "Translation 3"}
            self.client.post(self.xx_form_url, data)
            po_file_hash_after, mo_file_hash_after = (
                file_hash(po_file),
                file_hash(mo_file),
            )

            # Only the PO should have changed, the MO should be unchanged
            self.assertNotEqual(po_file_hash_before, po_file_hash_after)
            self.assertEqual(mo_file_hash_before, mo_file_hash_after)

            # Verify that translating another string also leaves the MO unchanged
            po_file_hash_before, mo_file_hash_before = (
                po_file_hash_after,
                mo_file_hash_after,
            )
            msg_hashes = message_hashes()
            data = {msg_hashes["String 2"]: "Translation 4"}
            self.client.post(self.xx_form_url, data)
            po_file_hash_after, mo_file_hash_after = (
                file_hash(po_file),
                file_hash(mo_file),
            )

            self.assertNotEqual(po_file_hash_before, po_file_hash_after)
            self.assertEqual(mo_file_hash_before, mo_file_hash_after)

        with self.settings(ROSETTA_AUTO_COMPILE=True):
            po_file_hash_before, mo_file_hash_before = (
                po_file_hash_after,
                mo_file_hash_after,
            )
            msg_hashes = message_hashes()
            data = {msg_hashes["String 2"]: "Translation 5"}
            self.client.post(self.xx_form_url, data)
            po_file_hash_after, mo_file_hash_after = (
                file_hash(po_file),
                file_hash(mo_file),
            )

            self.assertNotEqual(po_file_hash_before, po_file_hash_after)
            self.assertNotEqual(mo_file_hash_before, mo_file_hash_after)

    def test_41_pr_176_embed_in_admin(self):
        resp = self.client.get(reverse("admin:index"))
        self.assertContains(resp, "app-rosetta module")

    def _setup_view(self, view, request, *args, **kwargs):
        """Mimic as_view() returned callable, but returns view instance.

        args and kwargs are the same you would pass to ``reverse()``
        (From http://tech.novapost.fr/django-unit-test-your-views-en.html.)
        """
        view.request = request
        view.args = args
        view.kwargs = kwargs
        return view

    def test_42_view_property_po_file_is_writable(self):
        """Confirm that we're accurately determining the filesystem write-perms
        on our .po file.
        """
        self.copy_po_file_from_template("./django.po.template")

        # By default, we're writable
        request = RequestFactory().get(self.xx_form_url)
        request.user = self.user
        kwargs = {"po_filter": "third-party", "lang_id": "xx", "idx": 0}
        view = self._setup_view(
            view=views.TranslationFormView(), request=request, **kwargs
        )
        self.assertTrue(view.po_file_is_writable)

        # Now try again with the file not writable. (Regenerate the view, since
        # this po_file_is_writable is cached for the life of the request.)
        # make the pofile read-only
        os.chmod(self.dest_file, 292)  # 0444
        view = self._setup_view(
            view=views.TranslationFormView(), request=request, **kwargs
        )
        self.assertFalse(view.po_file_is_writable)

        # Cleanup
        os.chmod(self.dest_file, 420)  # 0644

    def test_43_view_property_po_file_path(self):
        """Confirm our class-based views properly parse/validate the path of the
        .po file in question derived from the url kwargs.
        """
        self.copy_po_file_from_template("./django.po.template")

        # By default, when all goes well, we get our existing .po file path
        request = RequestFactory().get(self.xx_form_url)
        request.user = self.user
        kwargs = {"po_filter": "third-party", "lang_id": "xx", "idx": 0}
        view = self._setup_view(
            view=views.TranslationFormView(), request=request, **kwargs
        )
        self.assertEqual(view.po_file_path, self.dest_file)

        # But if the language isn't an option, we get a 404
        with self.settings(
            ROSETTA_LANGUAGES=[lang for lang, __ in settings.LANGUAGES if lang != "xx"]
        ):
            view = self._setup_view(
                view=views.TranslationFormView(), request=request, **kwargs
            )
            with self.assertRaises(Http404):
                view.po_file_path

        # And if the index doesn't correspond with a file, we get a 404
        new_kwargs = {"po_filter": "third-party", "lang_id": "xx", "idx": 9}
        view = self._setup_view(
            view=views.TranslationFormView(),
            # Recycle request, even though url kwargs conflict with ones below.
            request=request,
            **new_kwargs,
        )
        with self.assertRaises(Http404):
            view.po_file_path

    def test_44_message_search_function(self):
        """Confirm that search of the .po file works across the various message
        fields.
        """
        self.copy_po_file_from_template("./django.po.test44.template")
        url = self.xx_form_url + "?query=%s"

        # Here's the message entry we're considering:
        # #. Translators: consectetur adipisicing
        # #: templates/eiusmod/tempor.html:43
        # msgid "Lorem ipsum"
        # msgstr "dolor sit amet"
        # It is buried at the end of the file, so without searching for it, it
        # shouldn't be on the page
        r = self.client.get(url % "")
        self.assertNotContains(r, "Lorem")

        # Search msgid
        r = self.client.get(url % "ipsum")
        self.assertContains(r, "Lorem")

        # Search msgstr
        r = self.client.get(url % "dolor")
        self.assertContains(r, "Lorem")

        # Search occurences
        r = self.client.get(url % "tempor")
        self.assertContains(r, "Lorem")

        # Search comments
        r = self.client.get(url % "adipisicing")
        self.assertContains(r, "Lorem")

        # Search context
        r = self.client.get(url % "pellentesque")
        self.assertContains(r, "Lorem")

    def test_45_issue186_plural_msg_search(self):
        """Confirm that search of the .po file works for plurals."""
        self.copy_po_file_from_template("./django.po.issue186.template")
        url = self.xx_form_url + "?query=%s"

        # Here's the message entry we're considering:
        # msgstr "%d Child"
        # msgid_plural "%d Childrenen"
        # msgstr[0] "%d Tchilt"
        # msgstr[1] "%d Tchildren"

        # First, confirm that we don't ALWAYS see this particular message on the
        # page.
        r = self.client.get(url % "kids")
        self.assertNotContains(r, "Child")

        # Search msgid_plural
        r = self.client.get(url % "childrenen")
        self.assertContains(r, "Child")

        # Search msgstr[0]
        r = self.client.get(url % "tchilt")
        self.assertContains(r, "Child")

        # Search msgstr[1]
        r = self.client.get(url % "tchildren")
        self.assertContains(r, "Child")

    def test_46_search_string_with_unicode_symbols(self):
        """Confirm that search works with unicode symbols"""
        url = self.xx_form_url + "?" + urlencode({"query": force_bytes("Лорем")})

        # It shouldn't raise
        r = self.client.get(url)
        self.assertEqual(r.status_code, 200)

    @vcr.use_cassette(
        "fixtures/vcr_cassettes/test_47_azure_ajax_translation.yaml",
        match_on=["method", "scheme", "host", "port", "path", "query", "raw_body"],
        record_mode="once",
    )
    @override_settings(DEEPL_AUTH_KEY=None, AZURE_CLIENT_SECRET="FAKE")
    def test_47_azure_ajax_translation(self):
        r = self.client.get(
            reverse("rosetta.translate_text") + "?from=en&to=fr&text=hello%20world"
        )
        self.assertContains(r, '"Salut tout le monde"')

    @vcr.use_cassette(
        "fixtures/vcr_cassettes/test_47_2_deeps_ajax_translation.yaml",
        match_on=["method", "scheme", "host", "port", "path", "query", "raw_body"],
        record_mode="once",
    )
    @override_settings(DEEPL_AUTH_KEY="FAKE", AZURE_CLIENT_SECRET=None)
    def test_47_2_deeps_ajax_translation(self):
        r = self.client.get(
            reverse("rosetta.translate_text") + "?from=en&to=fr&text=hello%20world"
        )
        self.assertContains(r, '"Salut tout le monde"')

    @override_settings(ROSETTA_REQUIRES_AUTH=True)
    def test_48_requires_auth_not_respected_issue_203(self):
        self.client.logout()
        r = self.client.get(self.all_file_list_url)
        self.assertRedirects(
            r,
            "{}?next=/rosetta/files/all/".format(settings.LOGIN_URL),
            fetch_redirect_response=False,
        )
        self.assertEqual(302, r.status_code)

    @override_settings(ROSETTA_REQUIRES_AUTH=False)
    def test_49_requires_auth_not_respected_issue_203(self):
        r = self.client.get(self.all_file_list_url)
        self.assertEqual(200, r.status_code)

    @override_settings(ROSETTA_REQUIRES_AUTH=True, ROSETTA_LOGIN_URL="/custom-url/")
    def test_50_custom_login_url(self):
        self.client.logout()
        r = self.client.get(self.all_file_list_url)
        self.assertRedirects(
            r, "/custom-url/?next=/rosetta/files/all/", fetch_redirect_response=False
        )
        self.assertEqual(302, r.status_code)

    def test_51_rosetta_languages(self):
        self.assertTrue("xx" in dict(settings.LANGUAGES))
        self.assertFalse("yy" in dict(settings.LANGUAGES))

        with self.settings(ROSETTA_LANGUAGES=(("xx", "foo language"),)):
            r = self.client.get(self.project_file_list_url)
            self.assertTrue("foo language" in r.content.decode())
            self.assertFalse("bar language" in r.content.decode())

        with self.settings(
            ROSETTA_LANGUAGES=(("xx", "foo language"), ("yy", "bar language"))
        ):
            r = self.client.get(self.project_file_list_url)
            self.assertTrue("foo language" in r.content.decode())
            self.assertTrue("bar language" in r.content.decode())

    def test_198_embed_in_admin_access_control(self):
        resp = self.client.get(reverse("admin:index"))
        self.assertContains(resp, "rosetta-content-main")

        with self.settings(ROSETTA_ACCESS_CONTROL_FUNCTION=lambda user: False):
            resp = self.client.get(reverse("admin:index"))
            self.assertNotContains(resp, "rosetta-content-main")

    @mock.patch("rosetta.poutil.os.path.exists")
    def test_273_override_case_sensitivity(self, path_mock):
        path_mock.exists.return_value = False
        # no setting
        find_pos("en")
        path_mock.assert_called_with(mock.ANY)

        path_mock.reset_mock()
        with override_settings(ROSETTA_CASE_SENSITIVE_FILESYSTEM=False):
            find_pos("en")
            path_mock.isfile.assert_not_called()

        path_mock.reset_mock()
        with override_settings(ROSETTA_CASE_SENSITIVE_FILESYSTEM=True):
            find_pos("en")
            path_mock.isfile.assert_not_called()


# Stubbed access control function
def no_access(user):
    return False



import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as test_command


class Tox(test_command):
    user_options = [("tox-args=", "a", "Arguments to pass to tox")]

    def initialize_options(self):
        test_command.initialize_options(self)
        self.tox_args = None

    def finalize_options(self):
        test_command.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import shlex

        import tox

        args = self.tox_args
        if args:
            args = shlex.split(self.tox_args)
        errno = tox.cmdline(args=args)
        sys.exit(errno)


with open("README.rst") as readme:
    long_description = readme.read()

setup(
    name="django-rosetta",
    version=__import__("rosetta").get_version(limit=3),
    description="A Django application that eases the translation of Django projects",
    long_description=long_description,
    author="Marco Bonetti",
    author_email="mbonetti@gmail.com",
    url="https://github.com/mbi/django-rosetta",
    license="MIT",
    packages=find_packages(exclude=["testproject", "testproject.*"]),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Software Development :: Localization",
        "Topic :: Software Development :: Internationalization",
        "Framework :: Django",
        "Framework :: Django :: 4.2",
        "Framework :: Django :: 5.0",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    include_package_data=True,
    zip_safe=False,
    install_requires=["Django >= 4.2", "requests >= 2.30.0", "polib >= 1.1.0"],
    tests_require=["tox", "vcrpy"],
    cmdclass={"test": Tox},
)



from django.apps import AppConfig

from .conf import settings as rosetta_settings


class RosettaAppConfig(AppConfig):
    name = "rosetta"

    def ready(self):
        from django.contrib import admin

        if rosetta_settings.SHOW_AT_ADMIN_PANEL:
            admin.site.index_template = "rosetta/admin_index.html"



import importlib

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from .conf import settings as rosetta_settings


def can_translate(user):
    return get_access_control_function()(user)


def get_access_control_function():
    """
    Return a predicate for determining if a user can
    access the Rosetta views
    """
    access_function = getattr(settings, "ROSETTA_ACCESS_CONTROL_FUNCTION", None)
    if access_function is None:
        return is_superuser_staff_or_in_translators_group
    elif isinstance(access_function, str):
        # Dynamically load a permissions function
        perm_module, perm_func = access_function.rsplit(".", 1)
        perm_module = importlib.import_module(perm_module)
        return getattr(perm_module, perm_func)
    elif callable(access_function):
        return access_function
    else:
        raise TypeError(access_function)


# Default access control test
def is_superuser_staff_or_in_translators_group(user):
    if not getattr(settings, "ROSETTA_REQUIRES_AUTH", True):
        return True
    try:
        if not user.is_authenticated:
            return False
        elif user.is_superuser and user.is_staff:
            return True
        else:
            return user.groups.filter(name="translators").exists()
    except AttributeError:
        if (
            not hasattr(user, "is_authenticated")
            or not hasattr(user, "is_superuser")
            or not hasattr(user, "groups")
        ):
            raise ImproperlyConfigured(
                "If you are using custom User Models you must implement a custom authentication method for Rosetta. See ROSETTA_ACCESS_CONTROL_FUNCTION here: https://django-rosetta.readthedocs.org/en/latest/settings.html"
            )
        raise


def can_translate_language(user, langid):
    try:
        if not rosetta_settings.ROSETTA_LANGUAGE_GROUPS:
            return can_translate(user)
        elif not user.is_authenticated:
            return False
        elif user.is_superuser and user.is_staff:
            return True
        else:
            return user.groups.filter(name="translators-%s" % langid).exists()

    except AttributeError:
        if (
            not hasattr(user, "is_authenticated")
            or not hasattr(user, "is_superuser")
            or not hasattr(user, "groups")
        ):
            raise ImproperlyConfigured(
                "If you are using custom User Models you must implement a custom authentication method for Rosetta. See ROSETTA_ACCESS_CONTROL_FUNCTION here: https://django-rosetta.readthedocs.org/en/latest/settings.html"
            )
        raise
